part of '../source_encoder.dart';

class ClassElementEncoder extends ElementEncoder<ClassElement> {
  static const ClassElementEncoder instance = ClassElementEncoder._();

  const ClassElementEncoder._() : super._(sdk.flutter_class);

  @override
  void load(Map<String, Element> elements, ClassElement element) {
    final id = identifier(element);
    if (elements.containsKey(id)) return;
    final loadType = _typeLoader(elements);
    elements[id] = element;
    element.supertype?.let(loadType);
    if (!acceptElement(element)) return;
    for (final typeParameter in element.typeParameters)
      TypeParameterElementEncoder.instance.load(elements, typeParameter);
    element.interfaces.forEach(loadType);
    for (final constructor in element.constructors)
      for (final parameter in constructor.parameters) ParameterElementEncoder.instance.load(elements, parameter);
  }

  @override
  String identifier(ClassElement element) => target.identifier(element.id);

  @override
  String reference(ClassElement element) {
    final libraryAlias = LibraryEncoder.libraryAliasOf(element);
    return libraryAlias.isEmpty ? element.name : '$libraryAlias.${element.name}';
  }

  @override
  String referenceDeclaration(ClassElement element) => reference(element);

  @override
  String variable(ClassElement element) => identifier(element);

  @override
  String declaration(ClassElement element) {
    final builder = DeclarationBuilder(kRawFlutterClass, identifier(element));
    final accept = acceptElement(element);
    final typeParameters = element.typeParameters;
    if (accept) builder.parameterToken(reference(element), typeParameters);
    builder.parameter(name: 'library', value: LibraryEncoder.libraryVariableOf(element));
    final exports = LibraryEncoder.exportsFor(element);
    if (exports.isNotEmpty) builder.parameter(name: 'exports', value: '{${exports.join(', ')}}');
    builder.parameter(name: 'name', value: '\'${element.name}\'');
    if (element.isAbstract) builder.parameter(name: 'isAbstract', value: 'true');
    if (accept && typeParameters.isNotEmpty)
      builder.parameterGetter(
          type: '$kSet<$kFlutterTypeParameter>',
          name: 'typeParameters',
          value: 'const {${typeParameters.map(TypeParameterElementEncoder.instance.variable).join(', ')}}');
    final superType = element.supertype;
    if (superType != null) {
      final variable = ClassTypeEncoder.instance.variable(superType);
      builder.parameterGetter(
          type: kFlutterClass,
          name: 'superType',
          value: superType.typeArguments.isNotEmpty ? 'const $variable' : variable);
    }
    final interfaces = element.interfaces;
    if (accept && interfaces.isNotEmpty)
      builder.parameterGetter(
          type: '$kSet<$kFlutterClass>',
          name: 'interfaces',
          value: 'const {${interfaces.map(ClassTypeEncoder.instance.variable).join(', ')}}');
    final constructors = element.constructors.where(acceptConstructor);
    if (accept && constructors.isNotEmpty)
      builder.parameterGetter(
          type: '$kSet<$kFlutterConstructor>',
          name: 'constructors',
          value: 'const {${constructors.map((constructor) => _encodeConstructor(builder, constructor)).join(', ')}}');
    final fields =
        CodegenConfig.generateFieldsFor(element) ? element.fields.where(acceptField) : const <FieldElement>[];
    if (accept && fields.isNotEmpty) {
      final fieldsVariable = sdk.flutter_parameter.identifier(element.id);
      builder.addDeclaration(
          'final $fieldsVariable = {${fields.map((field) => _encodeField(builder, field)).join(', ')}};');
      builder.parameterGetter(type: '$kSet<$kFlutterField>', name: 'fields', value: fieldsVariable);
    }
    return builder.build(variable: variable(element));
  }

  String _encodeConstructor(DeclarationBuilder parentBuilder, ConstructorElement constructor) {
    final declaring = constructor.enclosingElement;
    final builder = DeclarationBuilder(kFlutterConstructor, identifier(declaring));
    builder.parameter(name: 'declaring', value: variable(declaring));
    if (constructor.isConst) builder.parameter(name: 'isConst', value: 'true');
    final name = constructor.name;
    if (name.isNotEmpty) builder.parameter(name: 'name', value: '\'$name\'');
    final parameters = constructor.parameters.where(acceptParameter);
    if (parameters.isNotEmpty)
      builder.parameter(
          name: 'parameters', value: '{${parameters.map(ParameterElementEncoder.instance.declaration).join(', ')}}');
    final factoryVariable = sdk.flutter_constructor.identifier(constructor.id);
    builder.parameter(name: 'factory', value: factoryVariable);
    parentBuilder.addDeclaration(_encodeConstructorFactory(factoryVariable, constructor));
    DefaultValueResolver.consumeDeclarations().forEach(parentBuilder.addDeclaration);
    return builder.build();
  }

  String _encodeField(DeclarationBuilder parentBuilder, FieldElement field) {
    final declaring = field.enclosingElement as ClassElement;
    final builder = DeclarationBuilder(kFlutterField, identifier(declaring));
    final fieldType = field.type;
    final fieldName = field.name;
    var typeVariable = TypeEncoder.of(fieldType).variable(fieldType);
    if (fieldType is InterfaceType && fieldType.typeArguments.isNotEmpty) typeVariable = 'const $typeVariable';
    builder.parameter(name: 'type', named: false, value: typeVariable);
    builder.parameter(name: 'name', named: false, value: '\'$fieldName\'');
    builder.parameter(name: 'getter', named: false, value: '(${reference(declaring)} _) => _.$fieldName');
    if (!field.isFinal)
      builder.parameter(name: 'setter', named: false, value: '(${reference(declaring)} _, x) => _.$fieldName = x');
    return builder.build();
  }

  String _encodeConstructorFactory(String variable, ConstructorElement constructor) {
    final classElement = constructor.enclosingElement;
    final typeParameters = classElement.typeParameters;
    final inner = StringBuffer();
    if (constructor.isConst && constructor.parameters.isEmpty && typeParameters.isEmpty) inner.write('const ');
    inner.write(reference(classElement));
    if (typeParameters.isNotEmpty)
      inner.write('<${typeParameters.map(TypeParameterElementEncoder.instance.reference).join(', ')}>');
    if (constructor.name.isNotEmpty) inner.write('.${constructor.name}');
    inner.write('(');
    final parameters = constructor.parameters.where(acceptParameter).toList(growable: false);
    for (var i = 0; i < parameters.length; i++) {
      final parameter = parameters[i];
      if (i > 0) inner.write(', ');
      if (parameter.isNamed) inner.write('${ParameterElementEncoder.instance.reference(parameter)}: ');
      inner.write('p()');
    }
    inner.write(')');
    final body = encodeTokensUsage('c.tokens', typeParameters, inner.toString());
    return '$variable($kFactoryContext c, $kParameterReader p) => $body;';
  }
}
