part of '../source_encoder.dart';

class FunctionElementEncoder extends ElementEncoder<FunctionTypedElement> {
  static const FunctionElementEncoder instance = FunctionElementEncoder._();

  const FunctionElementEncoder._() : super._(sdk.flutter_function);

  @override
  void load(Map<String, Element> elements, FunctionTypedElement element) {
    final type = element.type;
    final id = identifier(element);
    if (elements.containsKey(id)) return;
    elements[id] = element;
    final loadType = _typeLoader(elements);
    loadType(type.returnType);
    if (!acceptElement(element)) return;
    for (final typeParameter in type.typeFormals) TypeParameterElementEncoder.instance.load(elements, typeParameter);
    for (final parameter in type.parameters) ParameterElementEncoder.instance.load(elements, parameter);
  }

  @override
  String identifier(FunctionTypedElement element) {
    final type = element.type;
    final typeParameterElementEncoder = ElementEncoder<TypeParameterElement>(sdk.flutter_type_parameter);
    final returnType = type.returnType.let((returnType) => TypeEncoder.of(returnType).identifier(returnType));
    final typeArgs = type.typeFormals.map(typeParameterElementEncoder.identifier).join();
    final parameters = type.parameters.map((parameter) {
      final parameterType = parameter.type;
      return TypeEncoder.of(parameterType).identifier(parameterType);
    }).join();
    return target.identifier(returnType, typeArgs, parameters);
  }

  @override
  String reference(FunctionTypedElement element) {
    final type = element.type;
    if (type.typeFormals.isNotEmpty) return kFunction;
    final out = StringBuffer();
    out.write(type.returnType.let((returnType) => TypeEncoder.of(returnType).reference(returnType)));
    out.write(' $kFunction(');
    final parameters = type.parameters.map((parameter) {
      final parameterType = parameter.type;
      return TypeEncoder.of(parameterType).reference(parameterType);
    }).join(', ');
    out.write(parameters);
    out.write(')');
    return out.toString();
  }

  @override
  String referenceDeclaration(FunctionTypedElement element) => reference(element);

  @override
  String variable(FunctionTypedElement element) => identifier(element);

  @override
  String declaration(FunctionTypedElement element) {
    final builder = DeclarationBuilder(kRawFlutterFunction, identifier(element));
    final type = element.type;
    final returnType = type.returnType;
    final typeFormals = type.typeFormals;
    final accept = acceptElement(element);
    if (accept && typeFormals.isEmpty)
      builder.parameterTokenFunction(
          reference(element), type.returnType, type.parameters.map((parameter) => parameter.type));
    builder.parameter(name: 'returnType', value: TypeEncoder.of(returnType).variable(returnType));
    if (accept) {
      if (typeFormals.isNotEmpty)
        builder.parameter(
            name: 'typeFormals',
            value: '{${typeFormals.map(TypeParameterElementEncoder.instance.variable).join(', ')}}');
      final parameters = type.parameters;
      if (parameters.isNotEmpty) {
        ParameterElementEncoder.resetParameterIndex();
        builder.parameter(
            name: 'parameters', value: '{${parameters.map(ParameterElementEncoder.instance.declaration).join(', ')}}');
      }
      final factoryVariable = sdk.flutter_function.identifier(identifier(element));
      builder.parameter(name: 'factory', value: factoryVariable);
      builder.addDeclaration(_encodeFunctionFactory(factoryVariable, type));
    }
    return builder.build(variable: variable(element));
  }

  String _encodeFunctionFactory(String variable, FunctionType function) {
    final returnType = function.returnType;
    final typeFormals = function.typeFormals;
    final parameters = function.parameters;
    final inner = StringBuffer();
    if (typeFormals.isNotEmpty)
      inner.write('<${typeFormals.map(TypeParameterElementEncoder.instance.referenceDeclaration).join(', ')}>');
    ParameterElementEncoder.resetParameterIndex();
    inner.write('(${parameters.map(ParameterElementEncoder.instance.referenceDeclaration).join(', ')}) => c.proxy(');
    inner.write('c, ');
    inner.write(typeFormals.isEmpty
        ? 'const [], '
        : '[${typeFormals.map((typeFormal) => '$kTypeToken<${TypeParameterElementEncoder.instance.reference(typeFormal)}>()').join(', ')}], ');
    ParameterElementEncoder.resetParameterIndex();
    inner.write(
        parameters.isEmpty ? 'const []' : '[${parameters.map(ParameterElementEncoder.instance.reference).join(', ')}]');
    inner.write(')');
    if (!returnType.isVoid && !returnType.isDynamic)
      inner.write(' as ${TypeEncoder.of(returnType).reference(returnType)}');
    final typeParameters = findTypeParameters([
      returnType,
      ...typeFormals.map((typeFormal) => typeFormal.bound).whereNotNull(),
      ...parameters.map((parameter) => parameter.type)
    ], typeFormals)
        .toSet()
        .toList(growable: false);
    final body = encodeTokensUsage('c.tokens', typeParameters, inner.toString());
    return '$variable($kFunctionContext c) => $body;';
  }
}
