// ignore_for_file: implementation_imports
import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/type.dart';
import 'package:analyzer/src/dart/element/element.dart';
import 'package:collection/collection.dart';
import 'package:flutterx_reflection_codegen/src/encoder/source_encoder.dart';
import 'package:flutterx_reflection_codegen/src/encoder/target_sdk.dart';
import 'package:flutterx_reflection_codegen/src/encoder/top_level/encoder_library.dart';
import 'package:flutterx_reflection_codegen/src/writer_framework.dart';

class CodegenConfig {
  static final Map<String, String> _defaultValues = {};
  static final Set<String> _includeFields = {};
  static final Set<String> _excludeFields = {};

  const CodegenConfig._();

  static void load(FrameworkWriter subject) {
    final sdkLibrary = subject.current;
    final configurationVariable = sdkLibrary.topLevelElements
        .whereType<TopLevelVariableElement>()
        .firstWhereOrNull((element) => element.name == 'configuration');
    final configuration = (configurationVariable as ConstVariableElement?)?.constantInitializer?.asMapNode();
    if (configuration == null)
      return print(
          'Configuration variable not found. To provide a custom configuration add \'const configuration = { ... };\'');
    final defaultValues = configuration['defaultValues']?.asMapNode();
    if (defaultValues != null) {
      final alias = LibraryEncoder.libraryAliasOf(sdkLibrary);
      for (final defaultValue in defaultValues.entries)
        CodegenConfig._defaultValues[defaultValue.key] = _computeDefaultValue(alias, defaultValue.value);
    }
    final includeFields = configuration['includeFields']?.asSetNode() ?? const <CollectionElement>[];
    for (final referenceId in includeFields)
      CodegenConfig._includeFields.add((referenceId as SimpleStringLiteral).value);
    final excludeFields = configuration['excludeFields']?.asSetNode() ?? const <CollectionElement>[];
    for (final referenceId in excludeFields)
      CodegenConfig._excludeFields.add((referenceId as SimpleStringLiteral).value);
    final generateFieldsCascade = configuration['generateFieldsCascade']?.asBool() ?? false;
    if (generateFieldsCascade) {
      final includeFields =
          subject.entries.whereType<ClassElement>().where((element) => _includeFields.contains(referenceId(element)));
      includeFields.forEach(_computeFieldsTree);
    }
  }

  static String? defaultValueFor(ParameterElement parameter) => _defaultValues[parameterReferenceId(parameter)];

  static bool generateFieldsFor(ClassElement element) => _includeFields.contains(referenceId(element));

  static void _computeFieldsTree(ClassElement element) {
    final included = <String>{};
    for (final field in element.fields.where(acceptField)) {
      final fieldType = field.type.element;
      if (fieldType is ClassElement && acceptElement(fieldType) && !fieldType.isEnum && !fieldType.library.isDartCore) {
        final id = referenceId(fieldType);
        final added = !_excludeFields.contains(id) && _includeFields.add(id);
        if (added) {
          _computeFieldsTree(fieldType);
          included.add(id);
        }
      }
    }
    if (included.isNotEmpty)
      print(
          'GenerateFieldsCascade: also included for field generation of ${element.name}:\n    ${included.join('\n    ')}');
  }

  static String _computeDefaultValue(String alias, Expression value) {
    if (value is Identifier) {
      final name = value.name;
      if (Identifier.isPrivateName(name))
        throw ArgumentError('Bad configuration syntax: default value identifier cannot be private, found $name');
      if (_isDefaultValueResolver(value.staticType))
        throw ArgumentError(
            'Bad configuration syntax: DefaultValueResolver must be declared directly inside \'defaultValues\' map, found $value');
      return '$alias.$name';
    }
    if (value is InstanceCreationExpression) {
      if (!_isDefaultValueResolver(value.constructorName.type.type))
        throw ArgumentError(
            'Bad configuration syntax: default values can only be identifiers or DefaultValueResolver, found $value');
      final arguments = value.argumentList.arguments;
      if (arguments.length != 1)
        throw ArgumentError(
            'Bad configuration syntax: DefaultValueResolver accepts 1 argument, found [${arguments.join(', ')}]');
      final identifier = arguments[0];
      if (identifier is! Identifier)
        throw ArgumentError(
            'Bad configuration syntax: DefaultValueResolver argument should be an identifier, found $identifier');
      return '$kDefaultValueResolver($alias.${identifier.name})';
    }
    throw ArgumentError(
        'Bad configuration syntax: default values can only be identifiers or DefaultValueResolver, found $value');
  }

  static bool _isDefaultValueResolver(DartType? type) {
    final element = type?.element;
    return element is ClassElement && element.name == 'DefaultValueResolver';
  }
}

extension _ConfigurationExt on Expression {
  Map<String, Expression> asMapNode() {
    final node = this;
    if (node is! SetOrMapLiteral) throw ArgumentError('Bad configuration syntax: expected map literal, found $node');
    final entries = node.elements.whereType<MapLiteralEntry>();
    return Map.fromEntries(entries.map((entry) {
      final key = entry.key;
      if (key is! SimpleStringLiteral) throw ArgumentError('Bad configuration syntax: keys must be string literals');
      return MapEntry(key.value, entry.value);
    }));
  }

  Set<CollectionElement> asSetNode() {
    final node = this;
    if (node is! SetOrMapLiteral) throw ArgumentError('Bad configuration syntax: expected set literal, found $node');
    return node.elements.toSet();
  }

  bool asBool() {
    final node = this;
    if (node is! BooleanLiteral) throw ArgumentError('Bad configuration syntax: expected bool literal, found $node');
    return node.value;
  }
}
