// ignore_for_file: implementation_imports
import 'package:_fe_analyzer_shared/src/scanner/token.dart';
import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/ast/syntactic_entity.dart';
import 'package:analyzer/dart/constant/value.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/nullability_suffix.dart';
import 'package:analyzer/dart/element/type.dart';
import 'package:analyzer/src/dart/element/element.dart';
import 'package:collection/collection.dart';
import 'package:dart_style/src/dart_formatter.dart';
import 'package:flutterx_reflection_codegen/extension/code.dart';
import 'package:flutterx_reflection_codegen/src/encoder/source_encoder.dart';
import 'package:flutterx_reflection_codegen/src/encoder/target_sdk.dart';
import 'package:flutterx_reflection_codegen/src/encoder/top_level/encoder_library.dart';
import 'package:flutterx_reflection_codegen/src/encoder/utils/codegen_configuration.dart';

class DefaultValueResolver {
  static final DartFormatter _formatter = DartFormatter(pageWidth: 500);
  static final Set<String> _defaultValueDeclarations = {};
  final ParameterElement parameter;
  final Map<String, LibraryElement> _aliases = {};
  final Map<String, Element> _scope = {};
  Expression? _defaultValue;

  DefaultValueResolver(this.parameter) {
    final impl = parameter as ParameterElementImpl;
    if (impl is ConstVariableElement) {
      _defaultValue = impl.constantInitializer;
      if (_defaultValue == null) return;
      final library = impl.enclosingUnit.library;
      _addToScope(library, const []);
      for (final import in library.imports) {
        final importAlias = import.prefix?.name;
        final importLibrary = import.importedLibrary;
        if (importLibrary == null) continue;
        if (importAlias == null) {
          _addToScope(importLibrary, import.combinators);
        } else {
          _aliases[importAlias] = importLibrary;
        }
      }
    }
  }

  String? resolve() {
    final userDefined = CodegenConfig.defaultValueFor(parameter);
    if (userDefined != null) return userDefined;
    final defaultValue = _defaultValue;
    if (defaultValue == null) return null;
    try {
      return _format(_parse(defaultValue));
    } catch (_) {}
    try {
      return _format(_stringify(parameter.computeConstantValue()));
    } catch (_) {}
    try {
      return _format(_provideDefaultValue(), resolveLint: false);
    } catch (_) {}
    print(
        'Cannot resolve default value $defaultValue for parameter key: \'${parameterReferenceId(parameter)}\'.\n    null will be used instead. This is unsafe because cannot reflect original behaviour of the underlying constructor\n    Please try defining manually a default value or a DefaultValueResolver for this key inside configuration.defaultValues map.');
  }

  static Set<String> consumeDeclarations() {
    final result = _defaultValueDeclarations.toSet();
    _defaultValueDeclarations.clear();
    return result;
  }

  String _parse(SyntacticEntity node) {
    final n = StringBuffer();
    if (node is PrefixedIdentifier) {
      final library = _aliases[node.prefix.name];
      final prefix = library == null ? _parse(node.prefix) : LibraryEncoder.libraryAliasOf(library);
      if (prefix.isNotEmpty) n.write('$prefix.');
      n.write(node.identifier.name);
    } else if (node is Identifier) {
      if (Identifier.isPrivateName(node.name)) throw StateError('Identifier is private: $node');
      n.write(_resolveIdentifier(node));
    } else if (node is NamedType) {
      final type = node.type?.element;
      if (type == null) throw StateError('Failed to resolve $node');
      final prefix = LibraryEncoder.libraryAliasOf(type);
      if (prefix.isNotEmpty) n.write('$prefix.');
      final identifier = node.name;
      final name = identifier is PrefixedIdentifier ? identifier.identifier.name : identifier.name;
      if (Identifier.isPrivateName(name)) throw StateError('Type is private: $node');
      n.write(name);
      node.typeArguments?.let(_parse).let(n.write);
      node.question?.let(_parse).let(n.write);
    } else if (node is ConstructorName) {
      n.write(_parse(node.type));
      final name = node.name?.name;
      if (name != null) {
        if (Identifier.isPrivateName(name)) throw StateError('Constructor is private: $node');
        n.write('.$name');
      }
    } else if (node is SimpleToken) {
      n.write(node.type.precedence == SELECTOR_PRECEDENCE ? node.lexeme : ' ${node.lexeme} ');
    } else if (node is ArgumentList) {
      n.write(node.leftParenthesis);
      n.write(node.arguments.map(_parse).join(', ').trim());
      n.write(node.rightParenthesis);
    } else if (node is TypeArgumentList) {
      n.write(node.leftBracket);
      n.write(node.arguments.map(_parse).join(', ').trim());
      n.write(node.rightBracket);
    } else if (node is ListLiteral) {
      final typeArguments = node.typeArguments;
      if (typeArguments != null) n.write(_parse(typeArguments));
      n.write(node.leftBracket);
      n.write(node.elements.map(_parse).join(', ').trim());
      n.write(node.rightBracket);
    } else if (node is AstNode) {
      n.write(node.childEntities.map(_parse).join());
    } else {
      n.write(node);
    }
    return n.toString();
  }

  String _resolveIdentifier(Identifier identifier) {
    final name = identifier.name;
    final resolved = _scope[name];
    if (resolved != null) {
      final alias = LibraryEncoder.libraryAliasOf(resolved);
      return alias.isEmpty ? name : '$alias.$name';
    }
    final classElement = parameter.thisOrAncestorOfType<ClassElement>();
    if (classElement != null) {
      final fields = classElement.fields.where((element) => element.isStatic);
      final methods = classElement.methods.where((element) => element.isStatic);
      if ([...fields, ...methods].any((element) => element.name == name)) {
        final className = classElement.name;
        final prefix = LibraryEncoder.libraryAliasOf(classElement);
        return prefix.isNotEmpty ? '$prefix.$className.$name' : '$className.$name';
      }
    }
    return name;
  }

  String _stringify(DartObject? value) {
    if (value == null || !value.hasKnownValue) throw StateError('Value of $value is unknown');
    if (value.isNull) return 'null';
    dynamic result;
    result = value.toBoolValue();
    if (result != null) return result.toString();
    result = value.toIntValue();
    if (result != null) return result.toString();
    result = value.toDoubleValue();
    if (result != null) return result.toString();
    result = value.toStringValue();
    if (result != null) return result.toString();
    result = value.toListValue();
    if (result != null) return '[${(result as List<DartObject>).map(_stringify).join(', ')}]';
    result = value.toSetValue();
    if (result != null) return '{${(result as Set<DartObject>).map(_stringify).join(', ')}}';
    result = value.toMapValue();
    if (result != null)
      return '{${(result as Map<DartObject?, DartObject?>).entries.map((entry) => '${_stringify(entry.key)}: ${_stringify(entry.value)}').join(', ')}}';
    result = value.toSymbolValue();
    if (result != null) return result;
    result = value.toTypeValue();
    if (result != null) return (result as DartType).getDisplayString(withNullability: true);
    final type = value.type;
    if (type is InterfaceType) {
      final typeElement = type.element;
      final name = typeElement.name;
      final libraryName = LibraryEncoder.libraryName(typeElement);
      final libraryAlias = LibraryEncoder.libraryAliasOf(typeElement);
      final referenceId = '$libraryName:$name';
      final reference = libraryAlias.isEmpty ? name : '$libraryAlias.$name';
      switch (referenceId) {
        case 'dart:ui:Color':
          return '$reference(0x${value.getField('value')?.toIntValue()?.toRadixString(16)})';
        case 'dart:core:Duration':
          return '$reference(microseconds: ${value.getField('_duration')?.toIntValue()})';
        case 'package:flutter/src/painting/edge_insets.dart:EdgeInsets':
          return '$reference.fromLTRB(${value.getField('left')?.toDoubleValue()}, ${value.getField('top')?.toDoubleValue()}, ${value.getField('right')?.toDoubleValue()}, ${value.getField('bottom')?.toDoubleValue()})';
      }
    }
    throw StateError('Could not resolve $value');
  }

  String _provideDefaultValue() {
    final out = StringBuffer();
    final constructor = parameter.enclosingElement as ConstructorElement;
    final classElement = constructor.enclosingElement;
    if (!classElement.fields.any((field) => field.name == parameter.name && field.type == parameter.type))
      throw StateError('parameter cannot be resolved ${classElement.name}.${parameter.name}');
    final variable = sdk.flutter_type_parameter.identifier(
        ClassElementEncoder.instance.identifier(classElement), ParameterElementEncoder.instance.identifier(parameter));
    out.write('$variable() => ${_mockConstructor(classElement, constructor, useConst: true)}.${parameter.name};');
    _defaultValueDeclarations.add(out.toString());
    return '$kDefaultValueResolver($variable)';
  }

  String _mockClass(ClassElement classElement, {required bool useConst}) {
    final constructors = classElement.constructors.where(acceptConstructor).toList(growable: false);
    if (constructors.isEmpty) throw StateError('Cannot mock type ${classElement.name} no constructor available');
    final result = <String>[];
    for (final constructor in constructors) {
      final isDefault = constructor.isDefaultConstructor;
      try {
        final value = _mockConstructor(classElement, constructor, useConst: useConst);
        if (isDefault) {
          result.clear();
          result.add(value);
          break;
        } else {
          result.add(value);
        }
      } catch (_) {}
    }
    if (result.isEmpty) throw StateError('Cannot mock type ${classElement.name}');
    return result.sorted((a, b) => a.length - b.length).first;
  }

  String _mockConstructor(ClassElement classElement, ConstructorElement constructor, {required bool useConst}) {
    final out = StringBuffer();
    final isConst = constructor.isConst;
    final canBeConst = !constructor.parameters.any(
        (parameter) => parameter.isNotOptional && (parameter.type is FunctionType || parameter.type.isDartCoreList));
    if (useConst && isConst && canBeConst) out.write('const ');
    final prefix = LibraryEncoder.libraryAliasOf(classElement);
    if (prefix.isNotEmpty) out.write('$prefix.');
    out.write(classElement.name);
    final typeParameters = classElement.typeParameters;
    if (typeParameters.isNotEmpty) {
      out.write('<${typeParameters.map((typeParameter) {
        final bound = typeParameter.bound?.element;
        if (bound == null) return kDynamic;
        final prefix = LibraryEncoder.libraryAliasOf(bound);
        return prefix.isNotEmpty ? '$prefix.${bound.name}' : bound.name;
      }).join(', ')}>');
    }
    if (constructor.name.isNotEmpty) out.write('.${constructor.name}');
    out.write('(');
    var usedParameters = 0;
    final constraints = _ParameterConstraints.compute(constructor).toList(growable: false);
    for (final parameter in constructor.parameters) {
      if (parameter.isNotOptional) {
        if (usedParameters++ > 0) out.write(', ');
        if (parameter.isNamed) out.write('${parameter.name}: ');
        final value = _mockValue(parameter.type,
            useConst: useConst && isConst && !canBeConst, constraints: _ParameterConstraints(parameter, constraints));
        out.write(_resolveLints(value, removeConst: false));
      }
    }
    out.write(')');
    return out.toString();
  }

  String _mockValue(DartType type, {bool useConst = false, _ParameterConstraints? constraints, int seed = 0}) {
    if (type is TypeParameterType) type = type.bound;
    final constPrefix = useConst ? 'const ' : '';
    if (type.isVoid || (type.nullabilitySuffix == NullabilitySuffix.question && constraints?.isNullable != false))
      return 'null';
    if (type.isDynamic || type.isDartCoreObject) return seed.toString();
    if (type.isDartCoreBool) return (constraints?.asBool == true).toString();
    if (type.isDartCoreInt || type.isDartCoreNum) return ((constraints?.asInt ?? 0) + seed).toString();
    if (type.isDartCoreDouble) return ((constraints?.asDouble ?? .0) + seed).toString();
    if (type.isDartCoreString) return constraints?.asString ?? '\'\'';
    if (type.isDartCoreList || type.isDartCoreIterable) {
      var componentType = (type as ParameterizedType).typeArguments.first;
      if (componentType is TypeParameterType) componentType = componentType.bound;
      var prefix = componentType.element?.let(LibraryEncoder.libraryAliasOf) ?? '';
      if (prefix.isNotEmpty) prefix += '.';
      return '$constPrefix<$prefix${componentType.getDisplayString(withNullability: true)}>[${Iterable.generate(constraints?.minLength ?? 0, (_) => _mockValue(componentType)).join(', ')}]';
    }
    if (type.isDartCoreMap) {
      final typeArguments = (type as ParameterizedType).typeArguments;
      var keyType = typeArguments[0];
      var valueType = typeArguments[1];
      if (keyType is TypeParameterType) keyType = keyType.bound;
      if (valueType is TypeParameterType) valueType = valueType.bound;
      var keyPrefix = keyType.element?.let(LibraryEncoder.libraryAliasOf) ?? '';
      if (keyPrefix.isNotEmpty) keyPrefix += '.';
      var valuePrefix = valueType.element?.let(LibraryEncoder.libraryAliasOf) ?? '';
      if (valuePrefix.isNotEmpty) valuePrefix += '.';
      return '$constPrefix<$keyPrefix${keyType.getDisplayString(withNullability: true)}, $valuePrefix${valueType.getDisplayString(withNullability: true)}>{${Iterable.generate(constraints?.minLength ?? 0, (i) => '${_mockValue(keyType, seed: i)}: ${_mockValue(valueType)}').join(', ')}}';
    }
    if (type.isDartCoreSet) {
      var componentType = (type as ParameterizedType).typeArguments.first;
      if (componentType is TypeParameterType) componentType = componentType.bound;
      var prefix = componentType.element?.let(LibraryEncoder.libraryAliasOf) ?? '';
      if (prefix.isNotEmpty) prefix += '.';
      return '$constPrefix<$prefix${componentType.getDisplayString(withNullability: true)}>{${Iterable.generate(constraints?.minLength ?? 0, (i) => _mockValue(componentType, seed: i)).join(', ')}}';
    }
    if (type is FunctionType) {
      var i = 0;
      return '(${type.parameters.map((_) => 'p${i++}').join(', ')}) => ${_mockValue(type.returnType, useConst: true)}';
    }
    final element = type.element;
    if (element is EnumElementImpl) {
      final prefix = LibraryEncoder.libraryAliasOf(element);
      return '${prefix.isEmpty ? '' : '$prefix.'}${element.name}.${element.constants.first.name}';
    }
    if (element is ClassElement) {
      switch (element.displayName) {
        case 'Widget':
          return '$constPrefix${LibraryEncoder.libraryAlias('package:flutter/src/widgets/basic.dart')}.SizedBox()';
        default:
          return _mockClass(element, useConst: useConst);
      }
    }
    throw StateError('Cannot mock type $type');
  }

  String _format(String string, {bool resolveLint = true}) {
    try {
      var result = _formatter.formatStatement('$string;');
      result = result.substring(0, result.length - 1);
      if (resolveLint) result = _resolveLints(result);
      return result;
    } catch (e) {
      print(e);
      rethrow;
    }
  }

  String _resolveLints(String source, {bool removeConst = true}) {
    // unnecessary_const
    const _const = 'const ';
    if (removeConst && source.startsWith(_const)) return _resolveLints(source.substring(_const.length));
    // use_full_hex_values_for_flutter_colors
    const _color = 'Color(';
    if (source.contains(_color)) {
      final ps = source.indexOf(_color) + _color.length;
      final pe = source.indexOf(')', ps);
      if (ps != -1 && pe != -1) {
        final color = source.substring(ps, pe);
        const _0x = '0x';
        if (!color.startsWith(_0x)) {
          final colorInt = int.tryParse(color);
          if (colorInt != null)
            return _resolveLints(
                '${source.substring(0, ps)}$_0x${colorInt.toRadixString(16).padLeft(8, '0')}${source.substring(pe)}');
        }
      }
    }
    // use_named_constants
    const _offset = 'Offset(0.0, 0.0)';
    if (source.contains(_offset)) {
      return _resolveLints(source.replaceFirst(_offset, 'Offset.zero'));
    }
    return source;
  }

  void _addToScope(LibraryElement library, List<NamespaceCombinator> combinators) {
    for (final export in library.exports) {
      final exportedLibrary = export.exportedLibrary;
      if (exportedLibrary != null) _addToScope(exportedLibrary, export.combinators);
    }
    final definedNames = Map.fromEntries(library.topLevelElements.map((element) => MapEntry(element.name!, element)));
    void _hide(List<String> hiddenNames) {
      for (final name in hiddenNames) {
        definedNames.remove(name);
        definedNames.remove('$name=');
      }
    }

    void _show(List<String> shownNames) {
      for (final name in shownNames) {
        var element = definedNames[name];
        if (element != null) definedNames[name] = element;
        final setterName = '$name=';
        element = definedNames[setterName];
        if (element != null) definedNames[setterName] = element;
      }
    }

    for (final combinator in combinators) {
      if (combinator is HideElementCombinator) {
        _hide(combinator.hiddenNames);
      } else if (combinator is ShowElementCombinator) {
        _show(combinator.shownNames);
      }
    }
    _scope.addAll(definedNames);
  }
}

/// Helps to very approximately provide values witch does not throw any exception in const constructor
class _ParameterConstraints {
  final ParameterElement parameter;
  final List<BinaryExpression> constraints;

  const _ParameterConstraints(this.parameter, this.constraints);

  static Iterable<BinaryExpression> compute(ConstructorElement constructor) sync* {
    final initializers = (constructor as ConstructorElementImpl).constantInitializers;
    Iterable<BinaryExpression> _explore(BinaryExpression expression) sync* {
      final lo = expression.leftOperand;
      final ro = expression.rightOperand;
      if (lo is BinaryExpression && ro is BinaryExpression) {
        yield* _explore(lo);
        yield* _explore(ro);
      } else {
        yield expression;
      }
    }

    for (final constraint in initializers.whereType<AssertInitializer>()) {
      final condition = constraint.condition;
      if (condition is BinaryExpression) yield* _explore(condition);
    }
  }

  bool get isNullable => constraints
      .map(_resolveConstraint)
      .whereNotNull()
      .none((condition) => condition.operator == BinaryOperator.notEqual && condition.operand == 'null');

  bool get asBool => constraints
      .map(_resolveConstraint)
      .whereNotNull()
      .none((condition) => condition.operator == BinaryOperator.notEqual && condition.operand == 'true');

  int? get asInt => asDouble?.toInt();

  double? get asDouble {
    for (final condition in constraints.map(_resolveConstraint).whereNotNull()) {
      final operand = double.tryParse(condition.operand);
      if (operand == null) continue;
      switch (condition.operator) {
        case BinaryOperator.equal:
        case BinaryOperator.lessOrEqual:
        case BinaryOperator.greaterOrEqual:
          return operand;
        case BinaryOperator.notEqual:
        case BinaryOperator.greater:
          return operand + 1;
        case BinaryOperator.less:
          return operand - 1;
      }
    }
    return null;
  }

  String? get asString {
    final length = minLength;
    if (length == null) return null;
    return ' ' * length;
  }

  int? get minLength {
    for (final condition in constraints.map(_resolveConstraint).whereNotNull()) {
      if (condition.subject != 'length') continue;
      final operand = int.tryParse(condition.operand);
      if (operand == null) continue;
      switch (condition.operator) {
        case BinaryOperator.equal:
        case BinaryOperator.greaterOrEqual:
          return operand;
        case BinaryOperator.greater:
          return operand + 1;
        case BinaryOperator.notEqual:
          if (operand == 0) return 1;
          break;
        case BinaryOperator.lessOrEqual:
        case BinaryOperator.less:
          break;
      }
    }
    return null;
  }

  static const _inverted = {
    BinaryOperator.equal: BinaryOperator.equal,
    BinaryOperator.notEqual: BinaryOperator.notEqual,
    BinaryOperator.greater: BinaryOperator.lessOrEqual,
    BinaryOperator.greaterOrEqual: BinaryOperator.less,
    BinaryOperator.less: BinaryOperator.greaterOrEqual,
    BinaryOperator.lessOrEqual: BinaryOperator.greater,
  };

  static const _operators = {
    '==': BinaryOperator.equal,
    '!=': BinaryOperator.notEqual,
    '>': BinaryOperator.greater,
    '>=': BinaryOperator.greaterOrEqual,
    '<': BinaryOperator.less,
    '<=': BinaryOperator.lessOrEqual,
  };

  _PC? _resolveConstraint(BinaryExpression constraint) {
    final o = _operators[constraint.operator.toString()];
    if (o == null) return null;
    final name = parameter.name;
    final lo = constraint.leftOperand;
    final ro = constraint.rightOperand;
    _PC? _resolve(Identifier identifier, BinaryOperator operator, Expression operand) {
      if (operand is Identifier) return null; // too complex
      if (identifier is PrefixedIdentifier && identifier.prefix.name == name)
        return _PC(identifier.identifier.name, operator, operand.toString());
      if (identifier.name == name) return _PC('', operator, operand.toString());
      return null;
    }

    if (lo is Identifier) return _resolve(lo, o, ro);
    if (ro is Identifier) return _resolve(ro, _inverted[o]!, lo);
    return null;
  }
}

class _PC {
  final String subject;
  final BinaryOperator operator;
  final String operand;

  const _PC(this.subject, this.operator, this.operand);
}

enum BinaryOperator { equal, notEqual, greater, greaterOrEqual, less, lessOrEqual }
