part of '../flutterx_reflection.dart';

/// Flutter parameter used by [FlutterConstructor] and [FlutterFunction]
///
/// [isNamed] if the parameter is named (e.g. {required String parameter})
/// [isOptional] if the parameter is either named or positional optional (e.g. {String parameter = ''} or [String parameter = ''])
/// [isNullable] if the parameter is nullable (e.g. String? parameter)
/// [name] of parameter (or 'p$index' for unnamed parameters)
/// [type] of parameter can be typed even using class or function type arguments (e.g. ParametrizedFlutterClass(sMap, [sString, sT]))
/// [defaultValue] instance of parameter default value
/// [toString] describes this parameter accurately through [_describeFlutterParameter] function
@immutable
class FlutterParameter {
  final bool isNamed;
  final bool isOptional;
  final bool isNullable;
  final String name;
  final FlutterType type;
  final dynamic _defaultValue;

  @visibleForTesting
  bool get defaultHasResolver => _defaultValue is DefaultValueResolver;

  dynamic get defaultValue => defaultHasResolver ? (_defaultValue as DefaultValueResolver).resolve() : _defaultValue;

  const FlutterParameter({
    this.isNamed = false,
    this.isOptional = false,
    this.isNullable = false,
    required this.name,
    required this.type,
    dynamic defaultValue,
  }) : _defaultValue = defaultValue;

  @override
  String toString() => _describeFlutterParameter(this);
}

/// Getter for default value of a parameter.
/// If a default value of a constructor parameter is not accessible, create an instance of this class
/// and pass it as [FlutterParameter.defaultValue] argument in constructor
///
/// [resolve] the actual getter of the default value. Must a static function that return the default value instance
class DefaultValueResolver {
  final ValueGetter<dynamic> resolve;

  const DefaultValueResolver(this.resolve);
}
