part of '../flutterx_reflection.dart';

/// Flutter type
/// Check following implementations:
/// - [FlutterTypeAlias]
/// - [FlutterClass]
/// - [FlutterDynamic]
/// - [FlutterEnum]
/// - [FlutterFunction]
/// - [FlutterTypeParameter]
/// - [FlutterVoid]
///
/// [token] holds the reference to flutter [Type] instance
/// [name] is usually the type name
/// [toString] describes the type accurately
/// [isAssignableFrom] checks if other is assignable to this
@immutable
abstract class FlutterType {
  TypeToken get token;

  String get name;

  @override
  String toString();

  bool isAssignableFrom(FlutterType other);

  /// Check if types are assignable to each other
  static bool equals(FlutterType? type, FlutterType? other) {
    if (type == null) return other == null;
    if (other == null) return false;
    return type.isAssignableFrom(other) && other.isAssignableFrom(type);
  }
}

/// Global type that can be identified uniquely (unlike function type and type parameters)
/// Check following implementations:
/// - [FlutterTypeAlias]
/// - [FlutterClass]
/// - [FlutterDynamic]
/// - [FlutterEnum]
/// - [FlutterVoid]
///
/// [library] is equivalent to type's import declaration (e.g. 'dart:core' or 'package:flutter/foundation.dart')
/// [exports] all known libraries that exports [library]
/// [referenceId] is used to describe the raw type uniquely through [DefinedType.generateReferenceId] function
@immutable
abstract class DefinedType extends FlutterType {
  String get library;

  Set<String> get exports;

  String get referenceId;

  /// Generate reference id of type
  ///
  /// [type] instance providing library and name
  /// [export] optional export library, if not specified [library] will be used
  /// [returns] referenceId composed by '$library:$name'
  static String generateReferenceId(DefinedType type, {String? export}) => '${export ?? type.library}:${type.name}';

  /// Get all reference ids of a type including both [library] and [exports]
  ///
  /// [type] instance providing library and name
  /// [returns] iterable of referenceId generated by [generateReferenceId]
  static Iterable<String> allReferenceIds(DefinedType type) sync* {
    yield type.referenceId;
    yield* type.exports.map((export) => generateReferenceId(type, export: export));
  }
}

/// Raw type that can have type parameters
/// Check following implementations:
/// - [RawFlutterTypeAlias]
/// - [RawFlutterClass]
/// - [RawFlutterFunction]
///
/// [typeParameters] of type or empty set if the type has no type parameters
/// [parametrize] this type and return a [ParametrizedType] having given typeArgs and this as base type
@immutable
abstract class RawType extends FlutterType {
  Set<FlutterTypeParameter> get typeParameters;

  ParametrizedType parametrize(List<FlutterType> typeArgs);
}

/// Parametrized type with actual type arguments
/// Check following implementations:
/// - [ParametrizedFlutterTypeAlias]
/// - [ParametrizedFlutterClass]
/// - [ParametrizedFlutterFunction]
///
/// [base] is the raw type without typeArgs
/// [typeArgs] of base type parameters (should have same length of [RawType.typeParameters])
@immutable
abstract class ParametrizedType extends RawType {
  RawType get base;

  List<FlutterType> get typeArgs;
}
