part of '../flutterx_reflection.dart';

/// Type alias (e.g. 'typedef ValueGetter<T> = T Function()')
///
/// [token] @see [FlutterType.token]
/// [library] @see [DefinedType.library]
/// [exports] @see [DefinedType.exports]
/// [name] @see [FlutterType.name]
/// [referenceId] @see [DefinedType.referenceId]
/// [typeParameters] @see [RawType.typeParameters]
/// [aliasedType] is the reflected type of this typeAlias (e.g. 'T Function()')
/// [toString] @see [FlutterType.toString]
/// [isAssignableFrom] if typeAlias isAssignableFrom. @see [FlutterType.isAssignableFrom]
/// [parametrize] @see [RawType.parametrize]
@immutable
abstract class FlutterTypeAlias implements FlutterType, DefinedType, RawType {
  @override
  TypeToken get token;

  @override
  String get library;

  @override
  Set<String> get exports;

  @override
  String get name;

  @override
  String get referenceId;

  @override
  Set<FlutterTypeParameter> get typeParameters;

  FlutterType get aliasedType;

  @override
  String toString();

  @override
  bool isAssignableFrom(FlutterType other);

  @override
  ParametrizedFlutterTypeAlias parametrize(List<FlutterType> typeArgs);
}

/// Raw implementation of [FlutterTypeAlias]
@immutable
class RawFlutterTypeAlias implements FlutterTypeAlias {
  @override
  TypeToken get token => aliasedType.token;

  @override
  final String library;

  @override
  final Set<String> exports;

  @override
  final String name;

  @override
  String get referenceId => DefinedType.generateReferenceId(this);

  @override
  final Set<FlutterTypeParameter> typeParameters;

  @override
  final FlutterType aliasedType;

  const RawFlutterTypeAlias({
    required this.library,
    this.exports = const {},
    required this.name,
    this.typeParameters = const {},
    required this.aliasedType,
  });

  @override
  String toString() => _describeFlutterType(this);

  @override
  bool isAssignableFrom(FlutterType other) => aliasedType.isAssignableFrom(other);

  @override
  ParametrizedFlutterTypeAlias parametrize(List<FlutterType> typeArgs) => ParametrizedFlutterTypeAlias(this, typeArgs);
}

/// Parametrized implementation of [FlutterTypeAlias]
///
/// [base] @see [ParametrizedType.base]
/// [token] already parametrized basing on aliasedType. @see [FlutterTypeAlias.token]
/// [typeArgs] @see [ParametrizedType.typeArgs]
/// [aliasedType] is already parametrized basing on actual typeArgs and base typeParameters. @see [FlutterTypeAlias.aliasedType]
@immutable
class ParametrizedFlutterTypeAlias implements FlutterTypeAlias, ParametrizedType {
  @override
  final RawFlutterTypeAlias base;

  @override
  TypeToken get token => aliasedType.token;

  @override
  String get library => base.library;

  @override
  Set<String> get exports => base.exports;

  @override
  String get name => base.name;

  @override
  String get referenceId => base.referenceId;

  @override
  Set<FlutterTypeParameter> get typeParameters => base.typeParameters;

  @override
  final List<FlutterType> typeArgs;

  @override
  FlutterType get aliasedType {
    final aliasedType = base.aliasedType;
    // it makes no sense but is possible (e.g. typedef F<T> = void Function())
    if (aliasedType is! ParametrizedType) return aliasedType;
    final actualTypeArgs = typeParameters.mapTo(typeArgs);
    return aliasedType.parametrize(aliasedType.typeArgs.resolve(actualTypeArgs));
  }

  const ParametrizedFlutterTypeAlias(this.base, this.typeArgs);

  @override
  String toString() => _describeFlutterType(this);

  @override
  bool isAssignableFrom(FlutterType other) => aliasedType.isAssignableFrom(other);

  @override
  ParametrizedFlutterTypeAlias parametrize(List<FlutterType> typeArgs) => ParametrizedFlutterTypeAlias(base, typeArgs);
}
