part of '../flutterx_reflection.dart';

/// Constructor of a class
///
/// [declaring] class of this constructor (always raw class)
/// [isConst] if constructor is constant
/// [name] of this constructor or empty string if unnamed
/// [parameters] of this constructor
/// [factory] used to create object instance
/// [toString] describes this constructor accurately through [_describeFlutterConstructor] function
/// [parameter] simply return parameter by name
/// [newInstance] create new instance of declaring class using given typeArgs
@immutable
class FlutterConstructor {
  final RawFlutterClass declaring;
  final bool isConst;
  final String name;
  final Set<FlutterParameter> parameters;
  final ConstructorFactory factory;

  const FlutterConstructor({
    required this.declaring,
    this.isConst = false,
    this.name = '',
    this.parameters = const {},
    required this.factory,
  });

  @override
  String toString() => _describeFlutterConstructor(this);

  FlutterParameter parameter(String name) => parameters.require((item) => item.name == name,
      message: () => 'parameter \'$name\' not found in constructor $this');

  dynamic newInstance({required InstanceProvider provider, List<FlutterType> typeArgs = const []}) {
    try {
      final actualTypeArgs = declaring.typeParameters.mapTo(typeArgs);
      final context = FactoryContext._(this, actualTypeArgs, typeArgs.tokens, provider);
      final reader = ParameterReader._(context);
      return factory(context, reader);
    } catch (e, stack) {
      throw InvocationTargetException._(e, stack, declaring, 'Failed to instantiate constructor $this');
    }
  }
}

/// newInstance implementation: creates object instance from context and reader
///
/// [context] of constructor you can use to access factory data
/// [read] should be called for every parameter of the constructor call in declaration order
/// [returns] instance object
typedef ConstructorFactory = dynamic Function(FactoryContext context, ParameterReader read);

/// Provide instance value of each constructor parameter
///
/// [provide] must provide the value of given parameter (even default value which can be accessed from [FlutterParameter.defaultValue])
abstract class InstanceProvider {
  const InstanceProvider();

  dynamic provide(ParameterContext context);
}

/// Context for creating object instance
///
/// [constructor] calling constructor used to access parameters
/// [actualTypeArgs] is a map you can use to resolve type parameters of class and get actual type arguments
/// [tokens] of class typeArgs to get [Type] instance or empty if raw
/// [provider] provides parameter value from [ParameterContext]
/// [actualTypeOf] simply return typeArg corresponding to typeParameter in this context
@immutable
class FactoryContext {
  final FlutterConstructor constructor;
  final Map<FlutterTypeParameter, FlutterType> actualTypeArgs;
  final List<TypeToken> tokens;
  final InstanceProvider provider;

  const FactoryContext._(this.constructor, this.actualTypeArgs, this.tokens, this.provider);

  FlutterType actualTypeOf(FlutterTypeParameter typeParameter) =>
      requireNotNull(actualTypeArgs[typeParameter], message: () => 'Failed to resolve type argument of $typeParameter');
}

/// Context for resolving parameter value
///
/// [parentContext] of constructor
/// [parameter] of this context
/// [actualTypeArgs] of parameter type
/// [actualType] of parameter already parametrized using actualTypeArgs
@immutable
class ParameterContext {
  final FactoryContext parentContext;
  final FlutterParameter parameter;

  List<FlutterType> get actualTypeArgs {
    final type = parameter.type;
    if (type is ParametrizedType) return type.typeArgs.resolve(parentContext.actualTypeArgs);
    if (type is RawType) return type.typeParameters.resolve(parentContext.actualTypeArgs);
    return const [];
  }

  FlutterType get actualType {
    final type = parameter.type;
    return type is RawType ? type.parametrize(actualTypeArgs) : type;
  }

  const ParameterContext._(this.parentContext, this.parameter);
}

/// Allow to read parameters in declaration order
///
/// [context] of constructor
/// [call] every time you call this function it returns the field instance value corresponding to the next parameter.
///        You can call this function as many times as constructor parameters length. The value is provided via
///        the [InstanceProvider] related to this context
@immutable
class ParameterReader {
  final FactoryContext context;
  final Iterator<FlutterParameter> _parameters;

  ParameterReader._(this.context) : _parameters = context.constructor.parameters.iterator;

  dynamic call() {
    if (!_parameters.moveNext())
      throw StateError('Illegal call to read: parameters are only ${context.constructor.parameters.length}');
    return context.provider.provide(ParameterContext._(context, _parameters.current));
  }
}
