part of '../flutterx_reflection.dart';

/// Function type (e.g. 'void Function<T>(List<T> values)')
///
/// [token] @see [FlutterType.token]
/// [name] @see [FlutterType.name]
/// [typeParameters] excluding typeFormals including type parameters used in/as returnType, parameters and typeFormals bound type  @see [findTypeParameters] @see [RawType.typeParameters]
/// [returnType] of this function
/// [typeFormals] type parameters of this generic function (e.g. <T>)
/// [parameters] of this function
/// [factory] used to create function instance
/// [toString] @see [FlutterType.toString]
/// [isAssignableFrom] if other is function and its returnType, typeFormals and parameters are assignable to this one. @see [FlutterType.isAssignableFrom]
/// [parametrize] @see [RawType.parametrize]
/// [parameter] simply return parameter by name (or 'p$index' for unnamed parameters)
/// [newInstance] create new instance of this function which will be implemented in [ProxyFunction]
@immutable
abstract class FlutterFunction implements FlutterType, RawType {
  @override
  TypeToken get token;

  @override
  String get name;

  @override
  Set<FlutterTypeParameter> get typeParameters;

  FlutterType get returnType;

  Set<FlutterTypeParameter> get typeFormals;

  Set<FlutterParameter> get parameters;

  FunctionFactory get factory;

  @override
  String toString();

  @override
  bool isAssignableFrom(FlutterType other);

  @override
  ParametrizedFlutterFunction parametrize(List<FlutterType> typeArgs);

  FlutterParameter parameter(String name);

  dynamic newInstance(ProxyFunction proxy);
}

/// Raw implementation of [FlutterFunction]
/// A generic function type can't be a type argument.
/// In this case we don't even need a TypeToken so we provide a default one
@immutable
class RawFlutterFunction implements FlutterFunction {
  @override
  final TypeToken token;

  @override
  String get name => toString();

  @override
  Set<FlutterTypeParameter> get typeParameters =>
      findTypeParameters([returnType, ...typeFormals, ...parameters.map((parameter) => parameter.type)], typeFormals)
          .toSet();

  @override
  final FlutterType returnType;

  @override
  final Set<FlutterTypeParameter> typeFormals;

  @override
  final Set<FlutterParameter> parameters;

  @override
  final FunctionFactory factory;

  const RawFlutterFunction({
    this.token = const TypeToken<Function>(),
    required this.returnType,
    this.typeFormals = const {},
    this.parameters = const {},
    required this.factory,
  });

  @override
  String toString() => _describeFlutterType(this);

  @override
  bool isAssignableFrom(FlutterType other) {
    if (other is FlutterTypeAlias) other = other.aliasedType; // handle type alias
    if (other is! FlutterFunction) return false; // different type
    if (!returnType.isAssignableFrom(other.returnType)) return false; // return type not assignable
    if (!typeFormals.areAssignableFrom(other.typeFormals)) return false; // different typeFormals parameters
    if (!other.parameters.map((p) => p.type).areAssignableFrom(parameters.map((p) => p.type)))
      return false; // parameters not assignable
    return true;
  }

  @override
  ParametrizedFlutterFunction parametrize(List<FlutterType> typeArgs) => ParametrizedFlutterFunction(this, typeArgs);

  @override
  FlutterParameter parameter(String name) => parameters.require((item) => item.name == name,
      message: () => 'parameter \'$name\' not found in constructor $this');

  @override
  dynamic newInstance(ProxyFunction proxy) {
    try {
      final context = FunctionContext._(this, const {}, const [], proxy);
      return factory(context);
    } catch (e, stack) {
      throw InvocationTargetException._(e, stack, this, 'Failed to instantiate function');
    }
  }
}

/// Parametrized implementation of [FlutterFunction]
///
/// [base] @see [ParametrizedType.base]
/// [token] already parametrized basing on typeArgs. @see [FlutterFunction.token]
/// [typeArgs] including typeArgs for [FlutterFunction.typeParameters] @see [ParametrizedType.typeArgs]
/// [isAssignableFrom] also check if typeArgs are assignable. @see [FlutterFunction.isAssignableFrom]
/// [newInstance] also implement typeArgs @see [FlutterFunction.newInstance]
class ParametrizedFlutterFunction implements FlutterFunction, ParametrizedType {
  @override
  final RawFlutterFunction base;

  @override
  TypeToken get token => base.token.parametrize(typeArgs.tokens);

  @override
  String get name => base.name;

  @override
  Set<FlutterTypeParameter> get typeParameters => base.typeParameters;

  @override
  final List<FlutterType> typeArgs;

  @override
  FlutterType get returnType => base.returnType;

  @override
  Set<FlutterTypeParameter> get typeFormals => base.typeFormals;

  @override
  Set<FlutterParameter> get parameters => base.parameters;

  @override
  FunctionFactory get factory => base.factory;

  const ParametrizedFlutterFunction(this.base, this.typeArgs);

  @override
  String toString() => _describeFlutterType(this);

  @override
  bool isAssignableFrom(FlutterType other) {
    if (other is FlutterTypeAlias) other = other.aliasedType; // handle type alias
    if (other is! ParametrizedFlutterFunction) return false; // other has no type args
    if (!base.isAssignableFrom(other)) return false; // different base type
    if (!typeArgs.areAssignableFrom(other.typeArgs)) return false; // type args not assignable
    return true;
  }

  @override
  ParametrizedFlutterFunction parametrize(List<FlutterType> typeArgs) => ParametrizedFlutterFunction(base, typeArgs);

  @override
  FlutterParameter parameter(String name) => base.parameter(name);

  @override
  dynamic newInstance(ProxyFunction proxy) {
    try {
      final actualTypeArgs = typeParameters.mapTo(typeArgs);
      final context = FunctionContext._(this, actualTypeArgs, typeArgs.tokens, proxy);
      return factory(context);
    } catch (e, stack) {
      throw InvocationTargetException._(e, stack, this, 'Failed to instantiate function');
    }
  }
}

/// newInstance implementation: creates function instance from context
///
/// [context] of function you can use to access factory data
/// [returns] instance function
typedef FunctionFactory = dynamic Function(FunctionContext context);

/// Context for creating function instance
///
/// [function] calling function used to access [ProxyFunction]
/// [actualTypeArgs] is a map you can use to resolve type parameters of class and get actual type arguments
/// [tokens] of function typeArgs to get [Type] instance or empty if raw
/// [proxy] provides function implementation
/// [actualTypeOf] simply return typeArg corresponding to typeParameter in this context
class FunctionContext {
  final FlutterFunction function;
  final Map<FlutterTypeParameter, FlutterType> actualTypeArgs;
  final List<TypeToken> tokens;
  final ProxyFunction proxy;

  const FunctionContext._(this.function, this.actualTypeArgs, this.tokens, this.proxy);

  FlutterType actualTypeOf(FlutterTypeParameter typeParameter) =>
      requireNotNull(actualTypeArgs[typeParameter], message: () => 'Failed to resolve type argument of $typeParameter');
}

/// Provides function implementation and return result
///
/// [context] of function you can use to access factory data
/// [typeArgs] tokens of function typeFormals
/// [parameters] values of the call
/// [returns] result of function call
typedef ProxyFunction = dynamic Function(FunctionContext context, List<TypeToken> typeArgs, List<dynamic> parameters);
