part of '../flutterx_reflection.dart';

/// TypeToken holds a reference to flutter [Type] instance
/// Also provide a method to parametrize this type dynamically
///
/// [T] type reference
/// [typeArgs] tokens of type args or empty if not yet parametrized
/// [hasTypeParameters] if type can have typeArgs
/// [hasTypeArgs] if type args is not empty (if false and hasTypeParameters is true it means is the raw type)
/// [toString] describe this type token and [T]
/// [use] allow to access [T] from outside this class through a function declared like '<T>() => 'HELLO $T'' and return result
/// [isInstance] check if an object is instance of [T]
/// [parametrize] if hasTypeParameters return a parametrized TypeToken having parametrized T and non empty typeArgs, else return this
@immutable
class TypeToken<T> {
  final ParametrizeTypeToken? _parametrize;
  final List<TypeToken> typeArgs;

  bool get hasTypeParameters => _parametrize != null;

  bool get hasTypeArgs => typeArgs.isNotEmpty;

  const TypeToken([this._parametrize, this.typeArgs = const []]);

  @override
  String toString() => 'TypeToken<$T>';

  dynamic use(Function use) => use<T>();

  bool isInstance(Object object) => object is T;

  TypeToken parametrize(List<TypeToken> typeArgs) => _parametrize?.call(typeArgs) ?? this;
}

/// Parametrize TypeToken (e.g. (typeArgs) => typeArgs[0].use(<E>() => TypeToken<Iterable<E>>()))
///
/// [typeArgs] tokens used to access types
/// [returns] parametrized TypeToken
typedef ParametrizeTypeToken = TypeToken Function(List<TypeToken> typeArgs);
