part of '../../flutterx_reflection.dart';

/// [mapTo] maps each [FlutterTypeParameter] to it's actual type basing on indexes
extension FlutterTypeParameterIterableExt on Iterable<FlutterTypeParameter> {
  Map<FlutterTypeParameter, FlutterType> mapTo(List<FlutterType> typeArgs) {
    if (length == typeArgs.length)
      return Map.fromEntries(mapIndexed((index, parameter) => MapEntry(parameter, typeArgs[index])));
    throw StateError(
        'typeArgs must have the same length of typeParameters but found typeParameters: {${join(', ')}} typeArgs: [${typeArgs.join(', ')}]');
  }
}

/// [tokens] simply maps tokens of this type collection
/// [resolve] type parameters contained in this iterable and map it to corresponding type of given map obtained from [FlutterTypeParameterIterableExt.mapTo]
/// [areAssignableFrom] check if all types of this are assignable to other
extension FlutterTypeIterableExt on Iterable<FlutterType> {
  List<TypeToken> get tokens => isEmpty ? const <TypeToken>[] : map((typeArg) => typeArg.token).toList(growable: false);

  List<FlutterType> resolve(Map<FlutterTypeParameter, FlutterType> actualTypeArgs) =>
      map((typeArg) => typeArg is FlutterTypeParameter ? actualTypeArgs[typeArg] ?? typeArg : typeArg)
          .toList(growable: false);

  bool areAssignableFrom(Iterable<FlutterType> other) =>
      length == other.length &&
      IterableZip<FlutterType>([this, other]).every((entry) => entry[0].isAssignableFrom(entry[1]));
}

/// Search every [FlutterTypeParameter] usage in given [types] except the ones contained in [exclude]
/// e.g. we have this function type 'Map<X, Y> Function<Z extends X>(Z p1)' and we call
///      'findTypeParameters([returnType, ...typeFormals, ...parameters.map((parameter) => parameter.type)], typeFormals)).toSet()'
///      result is {X, Y} without Z
///
/// [types] or type to search in
/// [exclude] collection of type parameters to exclude from result
/// [returns] collection of found typeParameters (can contain duplicated so just call [Iterable.toSet] to get distinct elements)
Iterable<FlutterTypeParameter> findTypeParameters(
    Iterable<FlutterType> types, Iterable<FlutterTypeParameter> exclude) sync* {
  for (final type in types) {
    if (type is FlutterTypeParameter) {
      final bound = type.bound;
      if (bound != null) yield* findTypeParameters([bound], exclude);
      if (!exclude.contains(type)) yield type;
    } else if (type is ParametrizedType) {
      yield* findTypeParameters(type.typeArgs, exclude);
    } else if (type is RawType) {
      yield* findTypeParameters(type.typeParameters, exclude);
    }
  }
}

/// Provide info about instantiation exception
///
/// [cause] throwable of this exception
/// [causeTrace] stack trace of cause
/// [type] where the exception has been thrown
/// [message] describing error
/// [toString] provide summary of this exception data
class InvocationTargetException implements Exception {
  final dynamic cause;
  final StackTrace causeTrace;
  final FlutterType type;
  final String message;

  const InvocationTargetException._(this.cause, this.causeTrace, this.type, this.message);

  @override
  String toString() =>
      'InvocationTargetException occurred in type \'$type\'\n$message. Caused by:\n    $cause\n$causeTrace';
}
